/*
 * A file to handle a bunch of categories.
 * @Shahzad Ali
 */

#include "category.h"
#include "categories.h"

/* Constructer for the frequency table */
categories::categories(int n, double r_th, long d_rate, char *outfile) {
  if (n < 1)
    return;
  num_cats = n+SPECIAL_CATEGORIES;
  next_free = 0;
  total_records = 0;
  if (r_th >= 0 || r_th <= 1)
    rare_threshold = r_th;
  
  /* Set the decay rate */
  if (d_rate > 0)
    decay_rate = d_rate;
  if (outfile){
    fout.open(outfile);
    if (!fout.good())
      cout << "Could not open output file for writing, using STDOUT" << endl;
  }
  
  Nr = Qn = prev_Qn = 0.0;
  
  frequency_table = new category*[n+SPECIAL_CATEGORIES];
  /* the +2 is for special categories "new" and "rare" */
  if (!frequency_table){
    cout << "Out of mem" << endl;
    exit (1);
  }
  for (int i=0; i < n; i ++){
    frequency_table[i] = new category(NULL, 0, 0);
    frequency_table[i]->invalidate();
  }
  /* The two special categories */
  frequency_table[n] = new category(NULL, 0, CAT_NEW);
  frequency_table[n]->validate;
  frequency_table[n+1] = new category(NULL, 0, CAT_RARE);
  frequency_table[n+1]->validate;
}

/* Freeing up memory properly */
categories::~categories(){
  for (int i=0; i < num_cats; i++ ) {
    delete frequency_table[i];
  }
}

/* Increase the value of Nr everytime this function is called
 * It is not just a simple increase.  A formula is used for it
 */
void categories::increment_Nr(long n){
  if (n < 1)
    return;
  double rate = -(log(0.5)/log(2))/(double)decay_rate;
  Nr = Nr + pow(2.0, (1.0-(double)n)*rate);
}

/* Return the next free category */
int categories::next_available() {
  if ((next_free + 1) >= (num_cats - SPECIAL_CATEGORIES))
    /* No more free categories */
    return -1;
  return next_free++;
}

/* Search for the index of a category */
int categories::index(char* n, short fl){
  if (fl == CAT_NEW)
    return num_cats-1;
  if (fl == CAT_RARE)
    return num_cats;

  for(int i=0; i< num_cats-SPECIAL_CATEGORIES; i++ ) {
    /* Only look at valid entries */
    if (!frequency_table[i]->isvalid())
      continue;
    if (!strcmp(frequency_table[i]->get_name(), n))
      return i;
  }
  return -1;
}

/* Search for a category and return it */
category* categories::get_category(char * n, short fl) {
  int i = index(n, fl);
  if (i < 0)
    return (category*) NULL;
  return frequency_table[i];
}

/* make a new category */
category* categories::new_category(char *n) {
  if (!*n) {
    cout << "Category has to have a name >= 1 character" << endl;
    return NULL;
  }
  int i = next_available();
  if (i < 0){
    cout << "Can not allocate a new category" << endl;
    cout << "Can only allocate " 
	 << (num_cats-SPECIAL_CATEGORIES) << " categories" << endl;
    exit(1);
  }
  frequency_table[i]->validate();
  frequency_table[i]->set_name(n, strlen(n));
  return frequency_table[i];
}

/* Read a file and make a frequency table of categories */
int categories::train(char* fn) {
  ifstream fin;
  char line[80];
  category* cat;

  if (!fn){
    cout << "Can not open NULL training datafile for reading" << endl;
    return (-1);
  }
  fin.open(fn);
  if (!fin.good()) {
    cout << "Can not open file " << fn << " for reading" << endl;
    return (-1);
  }
  while (!fin.eof()){
    fin >> line;
    
    if (!*line)
      continue;
    if ((cat = get_category(line, 0)) == NULL){
      cat = new_category(line);
      if (!cat)
	continue;
    }
    cat->increment();
    total_records++;
  }
}

/* Read the test file and try to calculate Qn at each step */
int categories::test(char* fn) {
  ifstream fin;
  char line[80];
  category* cat;
  int idx;
  long records_read = 0;

  if (!fn){
    cout << "Can not open NULL test datafile for reading" << endl;
    return (-1);
  }
  fin.open(fn);
  if (!fin.good()) {
    cout << "Can not open file " << fn << " for reading" << endl;
    return (-1);
  }
  while (!fin.eof()){
    fin >> line;

    if (!*line)
      continue;
    
    records_read ++;

    increment_Nr(records_read);
    
    idx = index(line, 0);
    if (idx >= 0)
      cat = frequency_table[idx];
    else {
      idx = index(NULL, CAT_RARE);
      cat = get_rare();
    }      
    for (int i=0; i<num_cats; i++){
      frequency_table[i]->calc_g_freq(Nr, (double)decay_rate, (idx == i));
    }
    calc_Qn();
    print_result(line, records_read);
  }
}

/* Print the total number of categories that are present in this test */
int categories::total_cats() {
  int res = 0;

  for (int i=0; i<num_cats; i++){
    if (!frequency_table[i]->isvalid() ||
	frequency_table[i]->isspecial())
      continue;
    res ++;
  }
  /* One more to the rare category! */
  return ++res;
}

/* Print the final result into the output stream */
void categories::print_result(char* n, long num){
  double deg_freedom = 0.0;

  deg_freedom = (total_cats() > 1)?(double)(total_cats()-1):1.0;
  if (!fout.good())
    cout << n << "\t" << Qn << "\t\t" << chi_square(Qn, deg_freedom) << endl;
  else
    fout << n << "\t" << Qn << "\t\t" << chi_square(Qn, deg_freedom) << endl;
}

/* Calculate the current value of Qn */
void categories::calc_Qn(){
  double sum=0.0;
  double cur_Qn=0.0;

  for (int i=0; i<num_cats; i++){
    if (!frequency_table[i]->isvalid() ||
	frequency_table[i]->isspecial())
      continue;
    if ((cur_Qn = frequency_table[i]->calc_Qn(Nr)) >= 0){
      sum += cur_Qn;
    }
    else {
      /* Qn was -1....WHY? */
      if (frequency_table[i]->get_freq() == 0){
	/* XXYY: Need to raise a flag here, */
	assert(0);
      }
    }
  }
  prev_Qn = Qn;
  cur_Qn = get_rare()->calc_Qn(Nr);
  if (cur_Qn < 0.0)
    cur_Qn = 0.0;
  Qn = sum + cur_Qn;
}

/* Calculate frequencies now that the numbers are calculated */
void categories::calc_frequencies(){
  for (int i=0; i< num_cats; i++) {
    if (!frequency_table[i]->isvalid() ||
	frequency_table[i]->isspecial())
      continue;
    frequency_table[i]->calc_freq(total_records);
  }
}

/* Get the rare category */
category* categories::get_rare() {
  /* NOTE: In this case, the last category is the rare category!!! */
  return frequency_table[num_cats-1];
}

/* Get the New category */
category* categories::get_new() {
  /* NOTE: In this case, the second last category is the new category!!! */
  return frequency_table[num_cats-2];
}

/* Function to calculate the probability using Chi-Square distr */
double categories::chi_square(double x, double n){
  double p = exp(-0.5*x);
  double k = n, a = n;
  double t;

  if (((int)n%2) == 1)
    p = p*sqrt(2.0*x/M_PI);

    while(k>=2.0) { 
      p=p*x/k; 
      k=k-2.0; 
    }
    t=p; 
    while(t>0.000001*p) { 
      a=a+2.0; 
      t=t*x/a; 
      p=p+t;
    }
    return 1.0-p;
}

/* Pack the table to make the rare category if needed */
void categories::pack() {
  double this_freq = 0.0;
  category* rare = get_rare();
  if (!rare)
    return;
  
  for (int i=0; i<num_cats; i++){
    if (!frequency_table[i]->isvalid() ||
	frequency_table[i]->isspecial())
      continue;
    if ((this_freq = frequency_table[i]->get_freq()) <= rare_threshold){
      /* This is a rare occurrence.  Lets merge it with the rare category */
      rare->set_fm(rare->get_fm()+frequency_table[i]->get_fm());
      rare->calc_freq(total_records);
      frequency_table[i]->invalidate();
    }
  }
}

/* Print the frequency table so far */
void categories::print_table(ostream & tout) {
  if (!tout)
    return;
  
  tout << "Categories\t#'s\tFrequency" << endl;

  for (int i=0; i < num_cats; i++) {
    if (!frequency_table[i]->isvalid() ||
	frequency_table[i]->isspecial())
      continue;
 
    tout << frequency_table[i]->get_name() << "\t\t" 
	 << frequency_table[i]->get_fm() << "\t"
	 << frequency_table[i]->get_freq()<< endl;
  } 
  /* Now lets print the special categories */

  category* rare = get_rare();
  if (rare)
    tout << "<Rare>\t\t" << rare->get_fm() 
	 << "\t" << rare->get_freq()<< endl;

  category* new_cat = get_new();
  if (new_cat)
    tout << "<New>\t\t" << new_cat->get_fm() << "\t" 
	 << new_cat->get_freq()<< endl;
}
